///////////////////////////////////////////////////////////////////////////////
//
//    G E N L I B  v0.99.22
//
///////////////////////////////////////////////////////////////////////////////
//
//  Genlib library
//  by Patrick Pelissier
//  ppelissier@hotmail.com
//  http://www.timetoteam.fr.st
//
//  Genlib C/nostub wrapper and auxiliary library
//  by Clem Vasseur of Icarus Productions
//  nitrogen@nrj.com
//  http://icarus.ticalc.org
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __GENLIB_H__
#define __GENLIB_H__

#ifdef USE_KERNEL
volatile int genlib__version02[0];
#else
#ifdef NO_EXIT_SUPPORT
#warning "Genlib needs exit support (NO_EXIT_SUPPORT removed)"
#undef NO_EXIT_SUPPORT
#endif
#ifdef SAVE_SCREEN
#warning "Genlib already saves the screen (SAVE_SCREEN removed)"
#undef SAVE_SCREEN
#endif
#endif

#include <tigcclib.h>
#if __TIGCC_MINOR__<94
#error "TIGCC v0.94 is required (http://tigcc.ticalc.org)"
#endif


// constants

#define GENLIB_VERSION "v0.99.22"

#define SCREEN_X      30    // horizontal screen size in bytes
#define SCREEN_Y      128   // vertical screen size in pixels
#define SCREEN_SIZE   (SCREEN_X*SCREEN_Y) // screen buffer size
#define DSCREEN_SIZE  (2*SCREEN_SIZE)     // dscreen buffer size

#define VIRTUAL_X     34    // horizontal vscreen size in pixels/8
#define VIRTUAL_Y     20    // vertical vscreen size in pixels/8
#define VSCREEN_SIZE  (VIRTUAL_X*VIRTUAL_Y*16)  // vscreen size

#define MAX_HANDLE    20  // number of available handles in gl_handle stack

#define MASTER        0   // main link_flag
#define SLAVE         1   // main link_flag
#define IN_PROGRESS   0   // link recpt/send flag
#define DONE          1   // link recpt/send flag


// macros

#define DARK_SCREEN(dscr)   ((SCREEN *)(dscr))
#define LIGHT_SCREEN(dscr)  (((SCREEN *)(dscr))+1)

#define ALLOC_SCREEN(scr) \
  char __tmp_##scr[SCREEN_SIZE+7]; \
  SCREEN *scr = (SCREEN *)(((long)__tmp_##scr+7)&(~7));

#define ALLOC_DSCREEN(dscr) \
  char __tmp_##dscr[DSCREEN_SIZE+7]; \
  DSCREEN *dscr = (DSCREEN *)(((long)__tmp_##dscr+7)&(~7));


// types

typedef unsigned char SCREEN[SCREEN_SIZE];    // 8x aligned
typedef unsigned char DSCREEN[DSCREEN_SIZE];  // 8x aligned
typedef short SPRITE_16[16][2];
typedef char TILE_8[8][2];
typedef short SPRITE_BW16[16];
typedef long FAST_SPRITE[16][16];
typedef char DHZ_TAB[SCREEN_Y];
typedef short HDZ_TAB[SCREEN_Y];
typedef char KEYBOARD[10];
typedef long FIXED;

typedef struct
{
  HANDLE handle;
  unsigned short size;
  unsigned char *table;
  void *tile;
  unsigned char *ind_tab; // internal use
  unsigned short xs_old;  // internal use
  unsigned short ys_old;  // internal use
  unsigned short xs;
  unsigned short ys;
  unsigned char VScreen[VSCREEN_SIZE];
} PLANE;

typedef struct
{
  unsigned short link_flag;
  unsigned char *buffer_recpt;
  unsigned char *buffer_send;
  unsigned short buffer_recpt_size;
  unsigned short buffer_send_size;
  unsigned char buffer_recpt_flag;
  unsigned char buffer_send_flag;
  unsigned short buffer_recpt_offset;
  unsigned short buffer_send_offset;
} LINK_DATA;

typedef struct
{
  char height;  // in pixels
  char weight;  // in words (pixels/16)
  short data[];
} BGS;

typedef struct
{
  unsigned int kh_key: 1;
  unsigned int kg_key: 1;
  unsigned int kf_key: 1;
  unsigned int ke_key: 1;
  unsigned int kd_key: 1;
  unsigned int kc_key: 1;
  unsigned int kb_key: 1;
  unsigned int ka_key: 1;
  unsigned int down_key: 1;
  unsigned int right_key: 1;
  unsigned int up_key: 1;
  unsigned int left_key: 1;
  unsigned int minus_key: 1;
  unsigned int plus_key: 1;
  unsigned int mode_key: 1;
  unsigned int exit_key: 1;
} JOYPAD;

typedef struct
{
  short x;
  short y;
} POINT;

typedef struct
{
  unsigned int tile: 10;
  unsigned int pal: 3;
  unsigned int flipv: 1;
  unsigned int fliph: 1;
  unsigned int user: 1;
} TILE;


// functions

#define gl_init                     genlib__0000
#define gl_quit                     genlib__0001
#define gl_init_dscreen             genlib__0002
#define gl_set_dscreen_int          genlib__0003
#define gl_set_dscreen_function     genlib__0004
#define gl_copy_window              genlib__0005
#define gl_init_plane               genlib__0006
#define gl_free_plane               genlib__0007
#define gl_refresh_plane            genlib__0008
#define gl_put_sprite_16            genlib__0009
//#define sprite_tile_adr           genlib__000A+2
//#define sprite_scr_x              genlib__000B+2
//#define sprite_scr_y              genlib__000C+2
//#define timer                     genlib__000D
#define gl_wait_no_key              genlib__000E
#define gl_wait_a_key               genlib__000F

#define gl_put_plane                genlib__0010
#define gl_put_sprite_16_flip_h     genlib__0011
#define gl_put_sprite_16_flip_v     genlib__0012
#define gl_put_sprite_16_flip_hv    genlib__0013
#define gl_put_big_sprite           genlib__0014
#define gl_put_big_sprite_flip_h    genlib__0015
#define gl_put_big_sprite_flip_v    genlib__0016
#define gl_put_big_sprite_flip_hv   genlib__0017
#define gl_put_sprite_16_last_flip  genlib__0018
#define gl_update_vscreen_8         genlib__0019
#define gl_key_pressed              genlib__001A
#define gl_copy_xyplane_xysprite    genlib__001B
#define gl_update_vscreen_16        genlib__001C
#define gl_exg_gray                 genlib__001D
#define gl_put_fgrd_plane           genlib__001E
#define gl_copy_dscreen_vscreen     genlib__001F

#define gl_put_plane_89             genlib__0020
#define gl_put_fgrd_plane_89        genlib__0021
#define gl_put_dhz_plane            genlib__0022
#define gl_put_dhz_fgrd_plane       genlib__0023
#define gl_make_fast_sprite         genlib__0024
#define gl_do_zoom                  genlib__0025
#define gl_put_sprite_16_zoom       genlib__0026
#define gl_put_big_sprite_zoom      genlib__0027
#define gl_cls                      genlib__0028
#define gl_draw_face                genlib__0029
#define gl_draw_c_face              genlib__002A
#define gl_put_pixel                genlib__002B
#define gl_set_byte_link            genlib__002C
#define gl_send_byte                genlib__002D
#define gl_receive_byte             genlib__002E
#define gl_send_data                genlib__002F

#define gl_synchronize              genlib__0030
//#define link_data                 genlib__0031
#define gl_set_LCD_MEM              genlib__0032
#define gl_set_link                 genlib__0033
#define gl_read_key_matrix          genlib__0034
#define gl_read_joypad              genlib__0035
#define gl_use_int1                 genlib__0036
//#define ?                         genlib__0037
#define gl_push_hd                  genlib__0038
#define gl_pop_hd                   genlib__0039
#define gl_free_hd                  genlib__003A
#define gl_put_dhz_plane_89         genlib__003B
#define gl_put_dhz_fgrd_plane_89    genlib__003C
#define gl_update_roll_vscreen_16   genlib__003D
#define gl_put_medium_string        genlib__003E
#define gl_put_small_string         genlib__003F

#define gl_put_large_string         genlib__0040
//#define window                    genlib__0041
//#define internal_timer            genlib__0042+3
#define gl_set_screen_int           genlib__0043
//#define hardware                  genlib__0044
#define gl_change_update            genlib__0045
#define gl_restore_update           genlib__0046
#define gl_put_dhz_trpt_plane       genlib__0047
#define gl_put_dhz_trpt_plane_89    genlib__0048
#define gl_init_screen              genlib__0049
#define gl_clear_window             genlib__004A
#define gl_draw_circle              genlib__004B
#define gl_draw_clipped_circle      genlib__004C
//#define flipping_tab              genlib__004D
#define gl_draw_clipped_face        genlib__004E
#define gl_draw_clipped_c_face      genlib__004F

#define gl_clip_line                genlib__0050
#define gl_draw_line                genlib__0051
#define gl_draw_clipped_line        genlib__0052
#define gl_set_filter               genlib__0053
//#define sin                       genlib__0054
//#define cos                       genlib__0055
#define gl_set_callback_update      genlib__0056
#define gl_free_callback_update     genlib__0057
#define gl_update_vscreen_max16     genlib__0058
#define gl_draw_bwhline_b           genlib__0059
#define gl_draw_bwhline_w           genlib__005A
#define gl_draw_hline_w             genlib__005B
#define gl_draw_hline_lg            genlib__005C
#define gl_draw_hline_dg            genlib__005D
#define gl_draw_hline_b             genlib__005E
#define gl_render_triangle          genlib__005F

#define gl_draw_disk                genlib__0060
#define gl_draw_clipped_disk        genlib__0061
#define gl_fill_screen              genlib__0062
#define gl_render_disk              genlib__0063
#define gl_draw_hline_light         genlib__0064
#define gl_create_bgs_string        genlib__0065
#define gl_pal_sprite_16            genlib__0066
//#define frame_timer               genlib__0067
#define gl_draw_hline_shadow        genlib__0068
#define gl_get_dscreen              genlib__0069
//#define version                   genlib__006A
#define gl_put_mask_spr16           genlib__006B
#define gl_put_mask_spr16_flip_h    genlib__006C
#define gl_put_mask_spr16_flip_v    genlib__006D
#define gl_put_mask_spr16_flip_hv   genlib__006E
#define gl_set_window               genlib__006F


// functions implemented in genlib.a

#ifndef USE_KERNEL
#undef gl_init
#undef gl_quit
#endif

#undef gl_render_triangle
#undef gl_put_dhz_plane
#undef gl_put_dhz_plane_89
#undef gl_put_dhz_fgrd_plane
#undef gl_put_dhz_fgrd_plane_89
#undef gl_do_zoom
#undef gl_clip_line

// functions without any argument

void gl_init(void);
void gl_quit(void);
void gl_exg_gray(void);
void gl_cls(void);
void gl_clear_window(void);
void gl_set_LCD_MEM(void);
void gl_set_window(void);
void gl_copy_window(void);
void gl_wait_no_key(void);
void gl_wait_a_key(void);
char gl_key_pressed(void);
KEYBOARD *gl_read_key_matrix(void);
JOYPAD gl_read_joypad(void);
void gl_set_byte_link(void);
char gl_receive_byte(void);
void gl_send_data(void);
char gl_synchronize(void);
void gl_free_hd(void);
HANDLE gl_pop_hd(void);
void gl_change_update(void);
void gl_restore_update(void);
DSCREEN *gl_get_dscreen(void);


// functions with register calling convention

#define D0 asm("d0")
#define D1 asm("d1")
#define D2 asm("d2")
#define D3 asm("d3")
#define D4 asm("d4")
#define D5 asm("d5")
#define D6 asm("d6")
#define D7 asm("d7")

#define A0 asm("a0")
#define A1 asm("a1")
#define A2 asm("a2")
#define A3 asm("a3")
#define A4 asm("a4")

void gl_set_filter(short filter D0);
void gl_send_byte(char byte D0);
void gl_push_hd(HANDLE Hd D0);
void gl_put_pixel(short x D0, short y D1, short color D3);
void gl_draw_circle(short x D4, short y D5, short radius D2, short color D3);
void gl_draw_clipped_circle(short x D4, short y D5, short radius D2, short color D3);
void gl_set_link(void* buf_rcpt A1, void* buf_send A2, short len_rcpt D1, short len_send D2);
void gl_put_small_string(SCREEN *scr A1, short x D0, short y D1, const char *str A0);
void gl_put_medium_string(SCREEN *scr A1, short x D0, short y D1, const char *str A0);
void gl_put_large_string(SCREEN *scr A1, short x D0, short y D1, const char *str A0);
void gl_put_big_sprite(short x D0, short y D1, BGS *bgs A0);
void gl_put_big_sprite_flip_h(short x D0, short y D1, BGS *bgs A1, BGS *tmp A0);
void gl_put_big_sprite_flip_v(short x D0, short y D1, BGS *bgs A1, BGS *tmp A0);
void gl_put_big_sprite_flip_hv(short x D0, short y D1, BGS *bgs A1, BGS *tmp A0);
void gl_put_sprite_16(short x D0, short y D1, short no D2);
void gl_put_sprite_16_flip_h(short x D0, short y D1, short no D2);
void gl_put_sprite_16_flip_v(short x D0, short y D1, short no D2);
void gl_put_sprite_16_flip_hv(short x D0, short y D1, short no D2);
void gl_set_screen_int(SCREEN *scr1 D0, SCREEN *scr2 D1, SCREEN *scr3 D2);
void gl_set_dscreen_int(DSCREEN *DScr D0);
void gl_set_dscreen_function(DSCREEN *DScr D0);
PLANE *gl_init_plane(unsigned char map[] A3, SPRITE_16 tile[] A4, short sizeX D3);
void gl_copy_dscreen_vscreen(PLANE *P A0, DSCREEN *Dscr A1);
void gl_copy_xyplane_xysprite(PLANE *P A0);
void gl_refresh_plane(PLANE *P A0);
void gl_free_plane(PLANE *P A0);
void gl_update_vscreen_8(PLANE *P A0);
void gl_update_vscreen_16(PLANE *P A0);
void gl_update_roll_vscreen_16(PLANE *P A0);
void gl_update_vscreen_max16(PLANE *P A0);
void gl_put_plane(PLANE *P A0);
void gl_put_plane_89(PLANE *P A0);
void gl_put_fgrd_plane(PLANE *P A0);
void gl_put_fgrd_plane_89(PLANE *P A0);
void gl_put_dhz_plane(PLANE *P A0, DHZ_TAB Dhz A4, HDZ_TAB Hdz A1);
void gl_put_dhz_plane_89(PLANE *P A0, DHZ_TAB Dhz A4, HDZ_TAB Hdz A1);
void gl_put_dhz_fgrd_plane(PLANE *P A0, DHZ_TAB Dhz A1, HDZ_TAB Hdz D7);
void gl_put_dhz_fgrd_plane_89(PLANE *P A0, DHZ_TAB Dhz A1, HDZ_TAB Hdz D7);
void gl_draw_face(POINT *Pt1 A1, POINT *Pt2 A2, POINT *Pt3 A3, SCREEN *scr A4);
void gl_draw_clipped_face(POINT *Pt1 A1, POINT *Pt2 A2, POINT *Pt3 A3, SCREEN *scr A4);
void gl_draw_c_face(POINT *Pt1 A1, POINT *Pt2 A2, POINT *Pt3 A3, short color D0);
void gl_draw_clipped_c_face(POINT *Pt1 A1, POINT *Pt2 A2, POINT *Pt3 A3, short color D0);
void gl_render_triangle(POINT *P1 A1, POINT *P2 A2, POINT *P3 A3, void (*draw_hline)(short x1 D0, short x2 D1, short y D3) A0);
void gl_draw_line(short x1 D0, short y1 D1, short x2 D2, short y2 D3, short color D5);
void gl_draw_clipped_line(short x1 D0, short y1 D1, short x2 D2, short y2 D3, short color D5);
void gl_draw_bwhline_b(short x1 D0, short x2 D1, short y D3);
void gl_draw_bwhline_w(short x1 D0, short x2 D1, short y D3);
void gl_draw_hline_w(short x1 D0, short x2 D1, short y D3);
void gl_draw_hline_lg(short x1 D0, short x2 D1, short y D3);
void gl_draw_hline_dg(short x1 D0, short x2 D1, short y D3);
void gl_draw_hline_b(short x1 D0, short x2 D1, short y D3);
void gl_draw_hline_light(short x1 D0, short x2 D1, short y D3);
void gl_draw_hline_shadow(short x1 D0, short x2 D1, short y D3);
void gl_draw_disk(short x D4, short y D5, short radius D2, short color D3);
void gl_draw_clipped_disk(short x D4, short y D5, short radius D2, short color D3);
void gl_render_disk(short x D4, short y D5, short radius D2, void (*draw_hline)(short x1 D0, short x2 D1, short y D3) A2);
void gl_fill_screen(SCREEN *scr A0, unsigned long pattern D0);
void gl_pal_sprite_16(short  pal D0, short src D1, short dest D2);
void gl_create_bgs_string(char *str A0, BGS *bgs A1);
void gl_make_fast_sprite(SPRITE_BW16 Spr A0, FAST_SPRITE fspr A1);
void gl_put_big_sprite_zoom(short x D0, short y D1, BGS *bgs A1, FIXED H_fact D4, FIXED V_fact D5);
void gl_put_sprite_16_zoom(short x D0, short y D1, short no D2, FIXED H_fact D4, FIXED V_fact D5);
void gl_put_tile_16(short x D0, short y D1, short no D2);
void gl_put_tile_16_flip_h(short x D0, short y D1, short n D2);
void gl_put_tile_16_flip_v(short x D0, short y D1, short n D2);
void gl_put_tile_16_flip_hv(short x D0, short y D1, short n D2);
void gl_do_zoom(short x, short y, FIXED H_fact, FIXED V_Fact, short hor_words, short vert_rows, SCREEN *scr, unsigned short *texture, short texture_inc);
int gl_clip_line(short *x1, short *y1, short *x2, short *y2);

#undef D0
#undef D1
#undef D2
#undef D3
#undef D4
#undef D5
#undef D6
#undef D7

#undef A0
#undef A1
#undef A2
#undef A3
#undef A4


// functions implemented as macros

/*
void gl_init_screen(SCREEN **scr, HANDLE *Hd);
void gl_init_dscreen(DSCREEN **scr, HANDLE *Hd);
void gl_set_spr_xy(short x, short y);
void gl_set_spr_tile(SPRITE_16 *tile_adr);
*/

#undef gl_init_screen
#define gl_init_screen(scr,hd) \
  asm("jsr genlib__0049; move.l %%a0,(%0); move.w %%d0,(%1)" :: "a" (scr), "a" (hd) : "d0", "d1", "d2", "a0", "a1")

#undef gl_init_dscreen
#define gl_init_dscreen(dscr,hd) \
  asm(" jsr genlib__0002; move.l %%a0,(%0); move.w %%d0,(%1)" :: "a" (dscr), "a" (hd) : "d0", "d1", "d2", "a0", "a1")

#undef gl_set_spr_xy
#ifdef USE_KERNEL
#define gl_set_spr_xy(x,y) \
  asm("move.w %0,genlib__000B+2; move.w %1,genlib__000C+2" :: "d" (x), "d" (y))
#else
#define gl_set_spr_xy(x,y) \
  asm("movea.l genlib__000B+2,%%a0; move.w %0,(%%a0); movea.l genlib__000C+2,%%a0; move.w %1,(%%a0)" :: "d" (x), "d" (y) : "a0")
#endif

#undef gl_set_spr_tile
#ifdef USE_KERNEL
#define gl_set_spr_tile(tile) \
  asm("move.l %0,genlib__000A+2" :: "a" (tile))
#else
#define gl_set_spr_tile(tile) \
  asm("movea.l genlib__000A+2,%%a0; move.l %0,(%%a0)" :: "a" (tile) : "a0")
#endif

// variables

#ifdef USE_KERNEL

#define gl_timer          genlib__000D
#define gl_link_data      genlib__0031
#define gl_window         genlib__0041
#define gl_hardware       genlib__0044
#define gl_flipping_tab   genlib__004D
#define gl_sin            genlib__0054
#define gl_cos            genlib__0055
#define gl_frame_timer    genlib__0067
#define gl_version        genlib__006A

extern volatile unsigned long gl_timer;
extern volatile LINK_DATA gl_link_data;
extern SCREEN *const gl_window;
extern const unsigned short gl_hardware;
extern const unsigned char gl_flipping_tab[256];
extern const signed char gl_sin[256];
extern const signed char gl_cos[256];
extern volatile unsigned short gl_frame_timer;
extern const char gl_version[];

#else

extern volatile unsigned long *const gl_timer;
extern volatile LINK_DATA *const gl_link_data;
extern SCREEN **const gl_window;
extern const unsigned short *const gl_hardware;
extern const unsigned char *const gl_flipping_tab;
extern const char *const gl_sin;
extern const char *const gl_cos;
extern volatile unsigned short *const gl_frame_timer;
extern const char *const gl_version;

#define gl_timer          (*gl_timer)
#define gl_link_data      (*gl_link_data)
#define gl_window         (*gl_window)
#define gl_hardware       (*gl_hardware)
#define gl_frame_timer    (*gl_frame_timer)

asm(
  " xdef gl_version\n"      "gl_version = genlib__006A+2\n"
  " xdef gl_window\n"       "gl_window = genlib__0041+2\n"
  " xdef gl_hardware\n"     "gl_hardware = genlib__0044+2\n"
  " xdef gl_flipping_tab\n" "gl_flipping_tab = genlib__004D+2\n"
  " xdef gl_sin\n"          "gl_sin = genlib__0054+2\n"
  " xdef gl_cos\n"          "gl_cos = genlib__0055+2\n"
  " xdef gl_timer\n"        "gl_timer = genlib__000D+2\n"
  " xdef gl_frame_timer\n"  "gl_frame_timer = genlib__0067+2\n"
  " xdef gl_link_data\n"    "gl_link_data = genlib__0031+2\n"
);

#endif


// GenLib AUXiliary library

// colors
enum { C_WHITE, C_LGRAY, C_DGRAY, C_BLACK, C_SHADOW=-3, C_LIGHT, C_NONE };

// text mode
enum { T_SMALL, T_MEDIUM, T_LARGE };

// functions
void glaux_rect(int x1, int y1, int x2, int y2, int fcolor, int bsize, int bcolor);
void glaux_circle(int x, int y, int r, int fcolor, int bcolor);
void glaux_triangle(int x1, int y1, int x2, int y2, int x3, int y3, int fcolor, int bcolor);
void glaux_text(int x, int y, int color, int textmode, const char *string, ...);
int glaux_read_char(void);
void *glaux_draw_hline_ptr(int color);

#endif
