#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "bin2oth.c"

int main(int argc, char *argv[])
{
    char*          infile    = NULL;
    char*          name      = NULL;
    int            calctype  = -1;
    FILE*          ifp;
    FILE*          ofp;
    char           outfile[1024];
    unsigned char* buffer;
    int            length;
    unsigned char* outbuffer = NULL;
    unsigned int   outlength;
    int            n;

  printf("\ngs2gen - Graphics Studio to Genlib Sprite16 converter\n");
  printf("written by Clem Vasseur <nitrogen@nrj.com>\n\n");

  if(argc == 2)
    infile = argv[1];
  else if(argc == 4)
  {
    infile = argv[1];
    if(!strcmp(argv[2], "-89"))      calctype  = CALC_TI89;
    else if(!strcmp(argv[2], "-92")) calctype  = CALC_TI92P;
    else
    {
      printf("Usage: gs2gen <infile> [<-89 | -92> <name>]\n");
      return 1;
    }
    name = argv[3];
  }
  else
  {
    printf("Usage: gs2gen <infile> [<-89 | -92> <name>]\n");
    return 1;
  }

  if (!(ifp = fopen(infile,"rb")))
  {
    fprintf(stderr,"ERROR: cannot open inputfile %s\n",infile);
    return 1;
  }

    // read infile into buffer
    fseek(ifp,0,SEEK_END);
    length = ftell(ifp) + 2;
    buffer = (unsigned char*)malloc(length);
    if(!buffer)
    {
        fprintf(stderr,"ERROR: cannot allocate memory (%d bytes)\n",length);
        fclose(ifp);
        return 1;
    }
    rewind(ifp);
    fread(buffer+2,1,length-2,ifp);
    fclose(ifp);

    buffer[0] = (length+4)/256;
    buffer[1] = (length+4)&255;

    if(!name) sprintf(outfile,"%s",infile);
    else if (calctype == CALC_TI89) sprintf(outfile,"%s.89y",name);
    else                       sprintf(outfile,"%s.9xy",name);

    if(!(ofp = fopen(outfile,"wb")))
    {
        fprintf(stderr,"ERROR: cannot open outputfile %s\n",outfile);
        free(buffer);
        return 1;
    }

    {
      int i,j;
      unsigned char buf[16*4];

      for(i=2; i<length;)
      {
        for(j=0; j<16*4; j++)
          buf[j] = buffer[i+j];
        for(j=0; j<32; j+=2)
        {
          buffer[i++] = buf[j];
          buffer[i++] = buf[j+1];
          buffer[i++] = buf[j+32];
          buffer[i++] = buf[j+33];
        }
      }
    }

    if(name)
      outbuffer = DataBuffer2OTHBuffer(calctype, NULL, name, "BIN", length-2, buffer+2, &outlength);
    else
    {
      outlength = length-2;
      outbuffer = (unsigned char *)malloc(outlength);
      if(!outbuffer)
      {
        fprintf(stderr,"ERROR: cannot allocate memory (%d bytes)\n",outlength);
        fclose(ofp);
        return 1;
      }
      memcpy(outbuffer, buffer+2, outlength);
    }

    n = 0;

    if (!outbuffer) {
        n = 1;
    }
    else {
        if (fwrite(outbuffer,outlength,1,ofp) != 1) {
            fprintf(stderr,"ERROR: cannot write %u bytes to outputfile to %s\n",outlength,outfile);
            n = 1;
        }
        else {
            fprintf(stderr,"%u bytes written to %s\n",outlength,outfile);
        }
        free(outbuffer);
    }

    free(buffer);
    fclose(ofp);

    return n;
}
