// void gl_init(void)
// void gl_quit(void)

#include <tigcclib.h>

#define GL_VERSION 2
#define fail(str) {ST_helpMsg(str); exit(0);}

void *genlib_buf;
void *bss_table;
//void (*ex)(void);
short calc_hw;
void *heap;

void genlib__0000(void);
void genlib__0001(void);
void genlib__000A(void);
void genlib__000B(void);
void genlib__000C(void);

short GetHardwareVersion(void)
{
  unsigned long hwpb, *rombase;
  rombase = (unsigned long *)((*(unsigned long *)0xC8) & 0x600000);
  hwpb = rombase[65];
  return (hwpb - (unsigned long)rombase  < 0x10000 &&
    *(unsigned short *)hwpb > 22 ? *(unsigned long *)(hwpb + 22) : 1);
}

static unsigned short *kernel_reloc(unsigned long addr, unsigned short *table)
{
  while(*table)
    *(unsigned long *)(genlib_buf + *table++) += addr;
  return table+1;
}

static unsigned short *kernel_reloc_w(unsigned short addr, unsigned short *table)
{
  while(*table)
    *(unsigned short *)(genlib_buf + *table++) += addr;
  return table+1;
}

void gl_init(void)
{
  SYM_ENTRY *sym;

  unsigned long program_counter;
  asm volatile("bsr 0f; 0:move.l (%%sp)+,%0":"=g"(program_counter));

  if(program_counter < 0x40000)
  {
    HANDLE tmp_h;
    unsigned char* tmp_src;
    unsigned short tmp_len;
    sym = SymFindFirst(NULL,2);

    while(sym)
    {
      tmp_h = sym->handle;
      if(tmp_h && HeapGetLock(tmp_h))
      {
        tmp_src = HeapDeref(tmp_h);
        tmp_len = *(unsigned short*)tmp_src+2;
        if((program_counter&0x3FFFF) > (unsigned long)tmp_src && (program_counter&0x3FFFF) < (unsigned long)tmp_src+tmp_len-2)
        {
          enter_ghost_space();
          EX_patch(tmp_src+0x40002,tmp_src+*(unsigned short*)tmp_src+0x40001);
          break;
        }
      }
      sym = SymFindNext();
    }
  }

  calc_hw = (TI89 ? 0:256) + GetHardwareVersion();

  short *p = (short *)HeapDeref;
  while(*p++ != 0x2078);
  heap = (void *)((long)(*p));

  for(sym=SymFindFirst(NULL,2); sym; sym=SymFindNext())
  {
    if(!strcmp(sym->name, "genlib") &&
       sym->handle && !sym->flags.bits.twin &&
       !(!sym->flags.bits.archived && HeapGetLock(sym->handle)))
    {
      HANDLE h = sym->handle;
      unsigned short *ptr = HeapDeref(h);
      unsigned short len = *ptr++;
      int i;

      ptr += 2;
      if(*ptr++!=0x3638 || *ptr++!=0x6B4C)
        fail("Genlib invalid");

      genlib_buf = malloc(len);
      if(!genlib_buf)
        fail("Out of memory");
      ptr = genlib_buf;
      memcpy(ptr, HeapDeref(h)+2, len);

      // save _exit entry point
      //ex = (void *)(ptr+ptr[0xE/2]/2);

      // check library version
      unsigned char *p = (unsigned char *)ptr;
      if(p[0x10] < GL_VERSION)
      {
        free(genlib_buf);
        fail("Genlib invalid");
      }

      // BSS relocations
      len = ptr[0x14/2]/2+1;
      bss_table = malloc(ptr[len++]);
      if(!bss_table)
      {
        free(genlib_buf);
        fail("Out of memory");
      }
      kernel_reloc((unsigned long)bss_table, ptr+len);

      // extract exported functions
      len = ptr[0x16/2]/2;
      i = ptr[len];
      if(i>0x70) i=0x70;
      void *func = genlib__0000 + 2;
      while(i--)
      {
        len++;
        *(unsigned long *)func = (unsigned long)(p + ptr[len]);
        func += 6;
      }

      // check extra RAM table and import table
      if(ptr[0x18/2] || ptr[0x1A/2])
      {
        free(genlib_buf);
        free(bss_table);
        fail("Genlib invalid");
      }

      // ROM_CALL relocations
      len = 0x1E/2;
      if(ptr[0x1C/2])
      {
        i = ptr[0x1E/2]+1;
        unsigned short *rom_table = ptr+0x1E/2+1;
        unsigned long rc;
        while(i--)
        {
          rc = *(unsigned long *)(*(long*)0xC8 + (*rom_table)*4);
          rom_table = kernel_reloc(rc, rom_table+1);
        }
        len = rom_table-ptr;
      }

      // RAM_CALL relocations
      if(ptr[len++])
      {
        i = ptr[len++]+1;
        unsigned short *ram_table = ptr+len;
        void *ram_ptr = NULL;
        while(i--)
        {
          switch(*ram_table & 0x00ff)
          {
            case 0:
              ram_ptr = &calc_hw;
              break;
            case 0xE: // font_medium
              ram_ptr = ROM_base;
              while(*(long *)ram_ptr != 0x54A854A8)
                ram_ptr += 2;
              break;
            case 0x11: // heap
              ram_ptr = heap;
          }
          if(*ram_table & 0x8000)
            ram_table = kernel_reloc_w((unsigned long)ram_ptr, ram_table+1);
          else
            ram_table = kernel_reloc((unsigned long)ram_ptr, ram_table+1);
        }
        len = ram_table-ptr;
      }

      kernel_reloc((unsigned long)p, ptr+len);

      (*(long *)(genlib__000A+2))+=2;
      (*(long *)(genlib__000B+2))+=2;
      (*(long *)(genlib__000C+2))+=2;

      genlib__0000();
      return;
    }
  }
  fail("Genlib not found");
}

void gl_quit(void)
{
  genlib__0001();
  //ex();
  WINDOW w;
  PortRestore();
  WinOpen(&w, MakeWinRect(0,0,239,127), 0x300);
  WinActivate(&w);
  WinClose(&w);
  DrawLine(0,LCD_HEIGHT-7,239,LCD_HEIGHT-7,A_NORMAL);
  ST_helpMsg(NULL);
  ST_eraseHelp();

  free(bss_table);
  free(genlib_buf);
}

asm(
  " xdef genlib__0000\n xdef genlib__0001\n xdef genlib__0002\n xdef genlib__0003\n"
  " xdef genlib__0004\n xdef genlib__0005\n xdef genlib__0006\n xdef genlib__0007\n"
  " xdef genlib__0008\n xdef genlib__0009\n xdef genlib__000A\n xdef genlib__000B\n"
  " xdef genlib__000C\n xdef genlib__000D\n xdef genlib__000E\n xdef genlib__000F\n"
  " xdef genlib__0010\n xdef genlib__0011\n xdef genlib__0012\n xdef genlib__0013\n"
  " xdef genlib__0014\n xdef genlib__0015\n xdef genlib__0016\n xdef genlib__0017\n"
  " xdef genlib__0018\n xdef genlib__0019\n xdef genlib__001A\n xdef genlib__001B\n"
  " xdef genlib__001C\n xdef genlib__001D\n xdef genlib__001E\n xdef genlib__001F\n"
  " xdef genlib__0020\n xdef genlib__0021\n xdef genlib__0022\n xdef genlib__0023\n"
  " xdef genlib__0024\n xdef genlib__0025\n xdef genlib__0026\n xdef genlib__0027\n"
  " xdef genlib__0028\n xdef genlib__0029\n xdef genlib__002A\n xdef genlib__002B\n"
  " xdef genlib__002C\n xdef genlib__002D\n xdef genlib__002E\n xdef genlib__002F\n"
  " xdef genlib__0030\n xdef genlib__0031\n xdef genlib__0032\n xdef genlib__0033\n"
  " xdef genlib__0034\n xdef genlib__0035\n xdef genlib__0036\n xdef genlib__0037\n"
  " xdef genlib__0038\n xdef genlib__0039\n xdef genlib__003A\n xdef genlib__003B\n"
  " xdef genlib__003C\n xdef genlib__003D\n xdef genlib__003E\n xdef genlib__003F\n"
  " xdef genlib__0040\n xdef genlib__0041\n xdef genlib__0042\n xdef genlib__0043\n"
  " xdef genlib__0044\n xdef genlib__0045\n xdef genlib__0046\n xdef genlib__0047\n"
  " xdef genlib__0048\n xdef genlib__0049\n xdef genlib__004A\n xdef genlib__004B\n"
  " xdef genlib__004C\n xdef genlib__004D\n xdef genlib__004E\n xdef genlib__004F\n"
  " xdef genlib__0050\n xdef genlib__0051\n xdef genlib__0052\n xdef genlib__0053\n"
  " xdef genlib__0054\n xdef genlib__0055\n xdef genlib__0056\n xdef genlib__0057\n"
  " xdef genlib__0058\n xdef genlib__0059\n xdef genlib__005A\n xdef genlib__005B\n"
  " xdef genlib__005C\n xdef genlib__005D\n xdef genlib__005E\n xdef genlib__005F\n"
  " xdef genlib__0060\n xdef genlib__0061\n xdef genlib__0062\n xdef genlib__0063\n"
  " xdef genlib__0064\n xdef genlib__0065\n xdef genlib__0066\n xdef genlib__0067\n"
  " xdef genlib__0068\n xdef genlib__0069\n xdef genlib__006A\n xdef genlib__006B\n"
  " xdef genlib__006C\n xdef genlib__006D\n xdef genlib__006E\n xdef genlib__006F\n"

  "genlib__0000: jmp 0.l\ngenlib__0001: jmp 0.l\ngenlib__0002: jmp 0.l\ngenlib__0003: jmp 0.l\n"
  "genlib__0004: jmp 0.l\ngenlib__0005: jmp 0.l\ngenlib__0006: jmp 0.l\ngenlib__0007: jmp 0.l\n"
  "genlib__0008: jmp 0.l\ngenlib__0009: jmp 0.l\ngenlib__000A: jmp 0.l\ngenlib__000B: jmp 0.l\n"
  "genlib__000C: jmp 0.l\ngenlib__000D: jmp 0.l\ngenlib__000E: jmp 0.l\ngenlib__000F: jmp 0.l\n"
  "genlib__0010: jmp 0.l\ngenlib__0011: jmp 0.l\ngenlib__0012: jmp 0.l\ngenlib__0013: jmp 0.l\n"
  "genlib__0014: jmp 0.l\ngenlib__0015: jmp 0.l\ngenlib__0016: jmp 0.l\ngenlib__0017: jmp 0.l\n"
  "genlib__0018: jmp 0.l\ngenlib__0019: jmp 0.l\ngenlib__001A: jmp 0.l\ngenlib__001B: jmp 0.l\n"
  "genlib__001C: jmp 0.l\ngenlib__001D: jmp 0.l\ngenlib__001E: jmp 0.l\ngenlib__001F: jmp 0.l\n"
  "genlib__0020: jmp 0.l\ngenlib__0021: jmp 0.l\ngenlib__0022: jmp 0.l\ngenlib__0023: jmp 0.l\n"
  "genlib__0024: jmp 0.l\ngenlib__0025: jmp 0.l\ngenlib__0026: jmp 0.l\ngenlib__0027: jmp 0.l\n"
  "genlib__0028: jmp 0.l\ngenlib__0029: jmp 0.l\ngenlib__002A: jmp 0.l\ngenlib__002B: jmp 0.l\n"
  "genlib__002C: jmp 0.l\ngenlib__002D: jmp 0.l\ngenlib__002E: jmp 0.l\ngenlib__002F: jmp 0.l\n"
  "genlib__0030: jmp 0.l\ngenlib__0031: jmp 0.l\ngenlib__0032: jmp 0.l\ngenlib__0033: jmp 0.l\n"
  "genlib__0034: jmp 0.l\ngenlib__0035: jmp 0.l\ngenlib__0036: jmp 0.l\ngenlib__0037: jmp 0.l\n"
  "genlib__0038: jmp 0.l\ngenlib__0039: jmp 0.l\ngenlib__003A: jmp 0.l\ngenlib__003B: jmp 0.l\n"
  "genlib__003C: jmp 0.l\ngenlib__003D: jmp 0.l\ngenlib__003E: jmp 0.l\ngenlib__003F: jmp 0.l\n"
  "genlib__0040: jmp 0.l\ngenlib__0041: jmp 0.l\ngenlib__0042: jmp 0.l\ngenlib__0043: jmp 0.l\n"
  "genlib__0044: jmp 0.l\ngenlib__0045: jmp 0.l\ngenlib__0046: jmp 0.l\ngenlib__0047: jmp 0.l\n"
  "genlib__0048: jmp 0.l\ngenlib__0049: jmp 0.l\ngenlib__004A: jmp 0.l\ngenlib__004B: jmp 0.l\n"
  "genlib__004C: jmp 0.l\ngenlib__004D: jmp 0.l\ngenlib__004E: jmp 0.l\ngenlib__004F: jmp 0.l\n"
  "genlib__0050: jmp 0.l\ngenlib__0051: jmp 0.l\ngenlib__0052: jmp 0.l\ngenlib__0053: jmp 0.l\n"
  "genlib__0054: jmp 0.l\ngenlib__0055: jmp 0.l\ngenlib__0056: jmp 0.l\ngenlib__0057: jmp 0.l\n"
  "genlib__0058: jmp 0.l\ngenlib__0059: jmp 0.l\ngenlib__005A: jmp 0.l\ngenlib__005B: jmp 0.l\n"
  "genlib__005C: jmp 0.l\ngenlib__005D: jmp 0.l\ngenlib__005E: jmp 0.l\ngenlib__005F: jmp 0.l\n"
  "genlib__0060: jmp 0.l\ngenlib__0061: jmp 0.l\ngenlib__0062: jmp 0.l\ngenlib__0063: jmp 0.l\n"
  "genlib__0064: jmp 0.l\ngenlib__0065: jmp 0.l\ngenlib__0066: jmp 0.l\ngenlib__0067: jmp 0.l\n"
  "genlib__0068: jmp 0.l\ngenlib__0069: jmp 0.l\ngenlib__006A: jmp 0.l\ngenlib__006B: jmp 0.l\n"
  "genlib__006C: jmp 0.l\ngenlib__006D: jmp 0.l\ngenlib__006E: jmp 0.l\ngenlib__006F: jmp 0.l\n"
);
